/*****************************************************************************
*       The Graphics Engine version 1.20                                     *
*                                                                            *
*       The Graphics Engine code and documentation are Copyright (c) 1993    *
*	by Matthew Hildebrand.                                               *
*                                                                            *
*       Unauthorised usage or modification of any or all of The Graphics     *
*       Engine is strictly prohibited.                                       *
*****************************************************************************/

/* This notice may not be removed and must remain intact */
static char copyright[] = "The Graphics Engine 1.20 -- Copyright (c) 1993 by Matthew Hildebrand";


/* Guess what... this file must be compiled with the large or huge model. */

#ifdef __TINY__
#error This module must be compiled with the large or huge model.
#endif
#ifdef __SMALL__
#error This module must be compiled with the large or huge model.
#endif
#ifdef M_I86SM
#error This module must be compiled with the large or huge model.
#endif
#ifdef __MEDIUM__
#error This module must be compiled with the large or huge model.
#endif
#ifdef M_I86MM
#error This module must be compiled with the large or huge model.
#endif
#ifdef __COMPACT__
#error This module must be compiled with the large or huge model.
#endif
#ifdef M_I86CM
#error This module must be compiled with the large or huge model.
#endif


#include <alloc.h>
#include <dos.h>
#include <math.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include "tge.h"

static void far TGE_deInitGraphics(void);
static void huge TGE_putImage(int x, int y, void far *image);
static void huge TGE_putImageInv(int x, int y, void far *image);
static void huge TGE_getImage(int ulx, int uly, int lrx, int lry, void far
	*image);
static void far TGE_putLine(int lineNum, int xOff, int lineLen, void far *buf);
static void far TGE_getLine(int lineNum, int xOff, int lineLen, void far *buf);
static unsigned long far TGE_imageSize(int ulx, int uly, int lrx, int lry);
static void far TGE_putPixel(int x, int y, unsigned colour);
static unsigned far TGE_getPixel(int x, int y);
static void far TGE_line(int x1, int y1, int x2, int y2, unsigned colour);
static void far TGE_horizLine(int y, int x1, int x2, unsigned colour);
static void far TGE_drawRect(int ulx, int uly, int lrx, int lry, unsigned
	colour);
static void far TGE_filledRect(int ulx, int uly, int lrx, int lry, unsigned
	colour);
static void far TGE_setPaletteReg(unsigned palReg, unsigned char red,
	unsigned char blue, unsigned char green);
static void far TGE_getPaletteReg(unsigned palReg, unsigned char *red,
	unsigned char *blue, unsigned char *green);
static void far TGE_setBlockPalette(unsigned firstReg, unsigned numRegs,
	void far *data);
static void far TGE_getBlockPalette(unsigned firstReg, unsigned numRegs,
	void far *data);
static void far TGE_clearGraphics(unsigned colour);
static void far TGE_ellipse(int xc, int yc, int wide, int deep, unsigned
	colour);
static void far TGE_filledEllipse(int xc, int yc, int wide, int deep, unsigned
	colour);
static void far TGE_circle(int x, int y, int radius, unsigned colour);
static void far TGE_filledCircle(int x, int y, int radius, unsigned colour);
static void far TGE_fillRegion(int x, int y, unsigned colour);
static unsigned far TGE_colourCloseTo(unsigned char red, unsigned char
	green, unsigned	char blue);
static unsigned far TGE_colourCloseToX(unsigned char red, unsigned char
	green, unsigned	char blue, unsigned colourExclude);

static void *forceOffset(void far *p);
static void setupInfo(struct GraphDrv far *grDrv);
static void near set4Pixels(int x, int y, int xc, int yc, unsigned colour);
static void near fill4Pixels(int x, int y, int xc, int yc, unsigned colour);


static void far *drvAddr = NULL;
struct GraphDrv far *TGEsys = NULL;


/******
******* Routines to load a graphics driver from disk.
******/

int loadGraphDriver(char *filename)
{
  struct GraphDrv far *grDrv = NULL;
  FILE *inFile;
  long fileLen;
  char signature[4];


  /* Force copyright string to be in EXE file */

  copyright[0] = 'T';				/* dummy write */


  /* Open the input file */

  if ((inFile=fopen(filename,"rb")) == NULL)	/* open the file */
    return (TGE_OPEN_ERR);


  /* Get file size */

  fseek(inFile, 0, SEEK_END);			/* seek to end of file */
  fileLen = ftell(inFile);			/* get current position */
  rewind(inFile);				/* seek to start of file */


  /* Parse the file header */

  if (!fread(signature, 1, 4, inFile))		/* read signature string */
  {
    fclose(inFile);
    return (TGE_FORMAT_ERR);
  }

  if (memcmp(signature,"TGE2",4))       	/* check signature string */
  {
    fclose(inFile);
    return (TGE_FORMAT_ERR);
  }


  /* Allocate memory for the driver */

  fileLen -= 4;					/* remove signature size */
  if (((void*)drvAddr=farmalloc(fileLen+15L)) == NULL)	/* grab some RAM */
  {
    fclose(inFile);
    return(TGE_ALLOC_ERR);
  }

  (void*)grDrv = forceOffset(drvAddr);		/* force to XXXX:0004 */


  /* Read in the body of the driver */

  if (fread(grDrv, 1, (unsigned)fileLen, inFile) != (unsigned)fileLen)
  {
    fclose(inFile);
    farfree(drvAddr);
    return (TGE_FILE_ERR);
  }


  /* Everything is loaded properly; now set up for program use */

  TGEsys = grDrv;				/* set global variable */
  setupInfo(grDrv);				/* setup function pointers */

  return (TGE_SUCCESS);				/* return success code */
}


/***
**** Force a pointer to the form XXXX:0004
***/
void *forceOffset(void far *p)
{
  void huge *temp = (void huge *)p;

  if (FP_OFF(temp) == 4)
    return ((void far *)temp);
  else if (FP_OFF(temp) > 4)
    return (MK_FP(FP_SEG(temp)+1, 4));
  else
    return (MK_FP(FP_SEG(temp), 4));
}


/***
**** Set up function pointers once a driver has been loaded
***/
void setupInfo(struct GraphDrv far *grDrv)
{
  register unsigned seg = FP_SEG(grDrv);

  /* initGraphics must always be present */
  (void *)grDrv->_initGraphics = MK_FP(seg, FP_OFF(grDrv->_initGraphics));

  /* These functions are optional; fill in the default ones if the
     driver doesn't contain them.  It's tedious, but somebody has
     to do it... */
  if (FP_OFF(grDrv->_deInitGraphics))	/* deInitGraphics */
    (void *)grDrv->_deInitGraphics = MK_FP(seg, FP_OFF(grDrv->_deInitGraphics));
  else
    grDrv->_deInitGraphics = TGE_deInitGraphics;

  if (FP_OFF(grDrv->_putImage))		/* putImage */
    (void *)grDrv->_putImage = MK_FP(seg, FP_OFF(grDrv->_putImage));
  else
    grDrv->_putImage = TGE_putImage;

  if (FP_OFF(grDrv->_putImageInv))	/* putImageInv */
    (void *)grDrv->_putImageInv = MK_FP(seg, FP_OFF(grDrv->_putImageInv));
  else
    grDrv->_putImageInv = TGE_putImageInv;

  if (FP_OFF(grDrv->_getImage))		/* getImage */
    (void *)grDrv->_getImage = MK_FP(seg, FP_OFF(grDrv->_getImage));
  else
    grDrv->_getImage = TGE_getImage;

  if (FP_OFF(grDrv->_putLine))		/* putLine */
    (void *)grDrv->_putLine = MK_FP(seg, FP_OFF(grDrv->_putLine));
  else
    grDrv->_putLine = TGE_putLine;

  if (FP_OFF(grDrv->_getLine))		/* getLine */
    (void *)grDrv->_getLine = MK_FP(seg, FP_OFF(grDrv->_getLine));
  else
    grDrv->_getLine = TGE_getLine;

  if (FP_OFF(grDrv->_imageSize))	/* imageSize */
    (void *)grDrv->_imageSize = MK_FP(seg, FP_OFF(grDrv->_imageSize));
  else
    grDrv->_imageSize = TGE_imageSize;

  if (FP_OFF(grDrv->_putPixel))		/* putPixel */
    (void *)grDrv->_putPixel = MK_FP(seg, FP_OFF(grDrv->_putPixel));
  else
    grDrv->_putPixel = TGE_putPixel;

  if (FP_OFF(grDrv->_getPixel))		/* getPixel */
    (void *)grDrv->_getPixel = MK_FP(seg, FP_OFF(grDrv->_getPixel));
  else
    grDrv->_getPixel = TGE_getPixel;

  if (FP_OFF(grDrv->_line))		/* line */
    (void *)grDrv->_line = MK_FP(seg, FP_OFF(grDrv->_line));
  else
    grDrv->_line = TGE_line;

  if (FP_OFF(grDrv->_horizLine))	/* horizLine */
    (void *)grDrv->_horizLine = MK_FP(seg, FP_OFF(grDrv->_horizLine));
  else
    grDrv->_horizLine = TGE_horizLine;

  if (FP_OFF(grDrv->_drawRect))		/* drawRect */
    (void *)grDrv->_drawRect = MK_FP(seg, FP_OFF(grDrv->_drawRect));
  else
    grDrv->_drawRect = TGE_drawRect;

  if (FP_OFF(grDrv->_filledRect))	/* filledRect */
    (void *)grDrv->_filledRect = MK_FP(seg, FP_OFF(grDrv->_filledRect));
  else
    grDrv->_filledRect = TGE_filledRect;

  if (FP_OFF(grDrv->_setPaletteReg))	/* setPaletteReg */
    (void *)grDrv->_setPaletteReg = MK_FP(seg, FP_OFF(grDrv->_setPaletteReg));
  else
    grDrv->_setPaletteReg = TGE_setPaletteReg;

  if (FP_OFF(grDrv->_getPaletteReg))	/* getPaletteReg */
    (void *)grDrv->_getPaletteReg = MK_FP(seg, FP_OFF(grDrv->_getPaletteReg));
  else
    grDrv->_getPaletteReg = TGE_getPaletteReg;

  if (FP_OFF(grDrv->_setBlockPalette))	/* setBlockPalette */
    (void *)grDrv->_setBlockPalette = MK_FP(seg, FP_OFF(grDrv->_setBlockPalette));
  else
    grDrv->_setBlockPalette = TGE_setBlockPalette;

  if (FP_OFF(grDrv->_getBlockPalette))	/* getBlockPalette */
    (void *)grDrv->_getBlockPalette = MK_FP(seg, FP_OFF(grDrv->_getBlockPalette));
  else
    grDrv->_getBlockPalette = TGE_getBlockPalette;

  if (FP_OFF(grDrv->_clearGraphics))	/* clearGraphics */
    (void *)grDrv->_clearGraphics = MK_FP(seg, FP_OFF(grDrv->_clearGraphics));
  else
    grDrv->_clearGraphics = TGE_clearGraphics;

  if (FP_OFF(grDrv->_ellipse))		/* ellipse */
    (void *)grDrv->_ellipse = MK_FP(seg, FP_OFF(grDrv->_ellipse));
  else
    grDrv->_ellipse = TGE_ellipse;

  if (FP_OFF(grDrv->_filledEllipse))	/* filledEllipse */
    (void *)grDrv->_filledEllipse = MK_FP(seg, FP_OFF(grDrv->_filledEllipse));
  else
    grDrv->_filledEllipse = TGE_filledEllipse;

  if (FP_OFF(grDrv->_circle))		/* circle */
    (void *)grDrv->_circle = MK_FP(seg, FP_OFF(grDrv->_circle));
  else
    grDrv->_circle = TGE_circle;

  if (FP_OFF(grDrv->_filledCircle))	/* filledCircle */
    (void *)grDrv->_filledCircle = MK_FP(seg, FP_OFF(grDrv->_filledCircle));
  else
    grDrv->_filledCircle = TGE_filledCircle;

  if (FP_OFF(grDrv->_fillRegion))	/* fillRegion */
    (void *)grDrv->_fillRegion = MK_FP(seg, FP_OFF(grDrv->_fillRegion));
  else
    grDrv->_fillRegion = TGE_fillRegion;

  if (FP_OFF(grDrv->_colourCloseTo))	/* colourCloseTo */
    (void *)grDrv->_colourCloseTo = MK_FP(seg, FP_OFF(grDrv->_colourCloseTo));
  else
    grDrv->_colourCloseTo = TGE_colourCloseTo;

  if (FP_OFF(grDrv->_colourCloseToX))	/* colourCloseToX */
    (void *)grDrv->_colourCloseToX = MK_FP(seg, FP_OFF(grDrv->_colourCloseToX));
  else
    grDrv->_colourCloseToX = TGE_colourCloseToX;
}


/******
******* Free graphics driver memory
******/

void unloadGraphDriver(void)
{
  if (drvAddr)
  {
    farfree(drvAddr);
    drvAddr = NULL;
  }
}


/******
******* High-level driver routines to be used if they aren't in a driver
******/

/***
**** deInitGraphics
***/
void TGE_deInitGraphics(void)
{
  _AX = 0x0003;
  geninterrupt(0x10);
}


/***
**** putImage
***/
void huge TGE_putImage(int x, int y, void far *image)
{
  register unsigned ycount;
  int wide, deep;
  register unsigned srcAdd = 0;
  unsigned char huge *p = (unsigned char *)image;

  wide = ((unsigned *)image)[0];	/* get image dimensions */
  deep = ((unsigned *)image)[1];
  p += 4;				/* point to start of image data */

  if (y < VIEWPORTULY)			/* clip y coordinate */
  {
    if (y+deep >= VIEWPORTULY)
    {
      p += wide * abs(VIEWPORTULY-y);	/* skip any offscreen rows */
      deep -= abs(VIEWPORTULY-y);	/* update image depth */
      y = VIEWPORTULY;			/* new y coordinate */
    }
    else
      return;
  }
  else if (y > VIEWPORTLRY)
    return;
  if (y+deep > VIEWPORTLRY)
    deep = VIEWPORTLRY-y + 1;

  if (x < VIEWPORTULX)			/* clip x coordinate */
  {
    if (x+wide >= VIEWPORTULX)
    {
      srcAdd += abs(VIEWPORTULX-x);
      p += abs(VIEWPORTULX-x);
      wide -= abs(VIEWPORTULX-x);
      x = VIEWPORTULX;
    }
    else
      return;
  }
  else if (x > VIEWPORTLRX)
    return;
  if (x+wide > VIEWPORTLRX)
  {
    srcAdd += wide-1 - (VIEWPORTLRX-x);
    wide = VIEWPORTLRX-x + 1;
  }

  if (wide && deep)	   		/* quit if offscreen */
  {
    srcAdd += wide;
    for (ycount=0; ycount<deep; ycount++)
    {
      putLine(y+ycount, x, wide, (void*)p);
      p += srcAdd;
    }
  }
}


/***
**** putImageInv
***/
void huge TGE_putImageInv(int x, int y, void far *image)
{
  register unsigned xcount, ycount;
  int wide, deep;
  register unsigned srcAdd = 0;
  unsigned char huge *p = (unsigned char *)image;

  wide = ((unsigned *)image)[0];	/* get image dimensions */
  deep = ((unsigned *)image)[1];
  p += 4;				/* point to start of image data */

  if (y < VIEWPORTULY)			/* clip y coordinate */
  {
    if (y+deep > VIEWPORTULY)
    {
      p += wide * abs(VIEWPORTULY-y);	/* skip any offscreen rows */
      deep -= abs(VIEWPORTULY-y);	/* update image depth */
      y = VIEWPORTULY;			/* new y coordinate */
    }
    else
      return;
  }
  else if (y > VIEWPORTLRY)
    return;
  if (y+deep > VIEWPORTLRY)
    deep = VIEWPORTLRY-y + 1;

  if (x < VIEWPORTULX)			/* clip x coordinate */
  {
    if (x+wide > VIEWPORTULX)
    {
      srcAdd += abs(VIEWPORTULX-x);
      p += abs(VIEWPORTULX-x);
      wide -= abs(VIEWPORTULX-x);
      x = VIEWPORTULX;
    }
    else
      return;
  }
  else if (x > VIEWPORTLRX)
    return;
  if (x+wide > VIEWPORTLRX)
  {
    srcAdd += wide-1 - (VIEWPORTLRX-x);
    wide = VIEWPORTLRX-x + 1;
  }

  if (wide && deep)	   		/* quit if offscreen */
  {
    for (ycount=0; ycount<deep; ycount++)
    {
      for (xcount=0; xcount<wide; xcount++)
      {
        if (*p)
          putPixel(x+xcount, y+ycount, *p);
        p++;
      }
      p += srcAdd;
    }
  }
}


/***
**** getImage
***/
void huge TGE_getImage(int ulx, int uly, int lrx, int lry, void far *image)
{
  register int ycount, wide, deep;
  int temp;
  unsigned char huge *p = (unsigned char*)image;

  if (lrx < ulx)			/* swap coordinates if necessary */
  {
    temp = ulx;
    lrx = ulx;
    lrx = temp;
  }
  if (lry < uly)
  {
    temp = uly;
    lry = uly;
    lry = temp;
  }

  if (ulx < VIEWPORTULX)		/* ensure coords are in bounds */
    ulx = VIEWPORTULX;
  else if (ulx > VIEWPORTLRX)
    ulx = VIEWPORTLRX;

  if (uly < VIEWPORTULY)
    uly = VIEWPORTULY;
  else if (uly > VIEWPORTLRY)
    uly = VIEWPORTLRY;

  if (lrx < VIEWPORTULX)
    lrx = VIEWPORTULX;
  else if (lrx > VIEWPORTLRX)
    lrx = VIEWPORTLRX;

  if (lry < VIEWPORTULY)
    lry = VIEWPORTULY;
  else if (lry > VIEWPORTLRY)
    lry = VIEWPORTLRY;

  ((unsigned *)image)[0] = wide = lrx-ulx+1;
  ((unsigned *)image)[1] = deep = lry-uly+1;
  p += 4;

  if (wide>0 && deep>0)
  {
    for (ycount=uly; ycount<=lry; ycount++)
    {
      getLine(ycount, ulx, wide, (void *)p);
      p += wide;
    }
  }
}


/***
**** putLine
***/
void TGE_putLine(int lineNum, int xOff, int lineLen, void far *buf)
{
  register unsigned count, y;

  for (count=0,y=lineNum; count<lineLen; count++)	/* draw the line */
    putPixel(xOff+count, y, *((unsigned char huge*)buf+count));
}


/***
**** getLine
***/
void TGE_getLine(int lineNum, int xOff, int lineLen, void far *buf)
{
  register unsigned count, y;

  for (count=0,y=lineNum; count<lineLen; count++)      	/* draw the line */
    *((unsigned char huge*)buf+count) = getPixel(xOff+count, y);
}


/***
**** imageSize
***/
unsigned long TGE_imageSize(int ulx, int uly, int lrx, int lry)
{
  unsigned long size;
  int temp;

  if (lrx < ulx)			/* swap coordinates if necessary */
  {
    temp = ulx;
    lrx = ulx;
    lrx = temp;
  }
  if (lry < uly)
  {
    temp = uly;
    lry = uly;
    lry = temp;
  }

  if (ulx < VIEWPORTULX)		/* ensure coords are in bounds */
    ulx = VIEWPORTULX;
  else if (ulx > VIEWPORTLRX)
    ulx = VIEWPORTLRX;

  if (uly < VIEWPORTULY)
    uly = VIEWPORTULY;
  else if (uly > VIEWPORTLRY)
    uly = VIEWPORTLRY;

  if (lrx < VIEWPORTULX)
    lrx = VIEWPORTULX;
  else if (lrx > VIEWPORTLRX)
    lrx = VIEWPORTLRX;

  if (lry < VIEWPORTULY)
    lry = VIEWPORTULY;
  else if (lry > VIEWPORTLRY)
    lry = VIEWPORTLRY;

  size = (unsigned long)(lrx-ulx+1);
  size *= (unsigned long)(lry-uly+1);

  return (size+4);			/* +4 for dimension information */
}


/***
**** putPixel
***/
void TGE_putPixel(int x, int y, unsigned colour)
{
  _AH = 0x0C;
  _AL = (unsigned char)colour;
  _BH = 0;
  _CX = x;
  _DX = y;
  geninterrupt(0x10);
}


/***
**** getPixel
***/
unsigned TGE_getPixel(int x, int y)
{
  _AH = 0x0D;
  _BH = 0;
  _CX = x;
  _DX = y;
  geninterrupt(0x10);

  _AH = 0;
  return (_AX);
}


/***
**** line
***/
void TGE_line(int x1, int y1, int x2, int y2, unsigned colour)
{
  register int t, distance;
  int xerr=0, yerr=0, deltax, deltay;
  int incx, incy;

  deltax = x2 - x1;			/* compute both distances */
  deltay = y2 - y1;

  if (deltax > 0)			/* compute increments */
    incx = 1;
  else if (deltax == 0)
    incx = 0;
  else
    incx = -1;

  if (deltay > 0)
    incy = 1;
  else if (deltay == 0)
    incy = 0;
  else
    incy = -1;

  deltax = abs(deltax);			/* determine greater distance */
  deltay = abs(deltay);
  if (deltax > deltay)
    distance = deltax;
  else
    distance = deltay;

  for (t=0; t<=distance+1; t++)		/* draw the line */
  {
    putPixel(x1, y1, colour);
    xerr += deltax;
    yerr += deltay;
    if (xerr > distance)
    {
      xerr -= distance;
      x1 += incx;
    }
    if (yerr > distance)
    {
      yerr -= distance;
      y1 += incy;
    }
  }
}


/***
**** horizLine
***/
void TGE_horizLine(int y, int x1, int x2, unsigned colour)
{
  register unsigned count, max;

  for (count=x1,max=x2; count<=max; count++)	/* draw the line */
    putPixel(count, y, colour);
}


/***
**** drawRect
***/
void TGE_drawRect(int ulx, int uly, int lrx, int lry, unsigned colour)
{
  register int ulxTemp, lrxTemp, temp;

  if (ulx > lrx)			/* swap coords if necessary */
  {
    temp = ulx;
    ulx = lrx;
    lrx = temp;
  }
  if (uly > lry)
  {
    temp = uly;
    uly = lry;
    lry = temp;
  }

  if (ulx<=VIEWPORTLRX && uly<=VIEWPORTLRY && lrx>=VIEWPORTULX &&
  	lry>=VIEWPORTULY)
  {
    ulxTemp = ulx>=VIEWPORTULX ? ulx : VIEWPORTULX;
    lrxTemp = lrx<=VIEWPORTLRX ? lrx : VIEWPORTLRX;

    if (uly >= VIEWPORTULY)
      horizLine(uly, ulxTemp, lrxTemp, colour);	/* top */
    if (lry <= VIEWPORTLRY)
      horizLine(lry, ulxTemp, lrxTemp, colour);	/* bottom */

    if (uly < VIEWPORTULY)		   	/* clip y-coordinates */
      uly = VIEWPORTULY;
    if (lry > VIEWPORTLRY)
      lry = VIEWPORTLRY;

    if (ulx >= VIEWPORTULX)
      line(ulx, uly, ulx, lry, colour);		/* left */
    if (lrx <= VIEWPORTLRX)
      line(lrx, uly, lrx, lry, colour);		/* right */
  }
}


/***
**** filledRect
***/
void TGE_filledRect(int ulx, int uly, int lrx, int lry, unsigned
	colour)
{
  register unsigned count, max;
  int temp;

  if (ulx > lrx)			/* swap coords if necessary */
  {
    temp = ulx;
    ulx = lrx;
    lrx = temp;
  }
  if (uly > lry)
  {
    temp = uly;
    uly = lry;
    lry = temp;
  }

  max = lry;
  for (count=uly; count<=max; count++)
    horizLine(count, ulx, lrx, colour);
}


/***
**** setPaletteReg
***/
void TGE_setPaletteReg(unsigned palReg, unsigned char red, unsigned char
	blue, unsigned char green)
{
  unsigned char r, g, b;

  r = red >> 2;				/* translate 8-bit to 6-bit */
  if ((red&3) >= 2)			/* round up if necessary */
    r++;
  if (r > 63)				/* ensure in range 0..63 */
    r = 63;

  g = green >> 2;
  if ((green&3) >= 2)
    g++;
  if (g > 63)
    g = 63;

  b = blue >> 2;
  if ((blue&3) >= 2)
    b++;
  if (b > 63)
    b = 63;

  _AX = 0x1010;				/* set up registers */
  _BX = palReg;
  _DH = r;
  _CH = g;
  _CL = b;
  geninterrupt(0x10);			/* call the video BIOS */
}


/***
**** getPaletteReg
***/
void TGE_getPaletteReg(unsigned palReg, unsigned char *red, unsigned char
	*blue, unsigned char *green)
{
  unsigned char r, g, b;

  _AX = 0x1015;				/* set up registers */
  _BX = palReg;
  geninterrupt(0x10);			/* call the video BIOS */
  r = _DH;
  g = _CH;
  b = _CL;

  *red = r << 2;			/* translate 6-bit to 8-bit */
  *green = g << 2;			/* and store new values */
  *blue = b << 2;
}


/***
**** setBlockPalette
***/
void TGE_setBlockPalette(unsigned firstReg, unsigned numRegs, void far *data)
{
  register unsigned count;
  unsigned char r, g, b;
  unsigned char huge *p;

  p = (unsigned char*) data;		/* initialize pointer */

  for (count=firstReg; count<=numRegs; count++)	/* main loop */
  {
    r = *p >> 2;			/* translate 8-bit to 6-bit */
    if ((*p&3) >= 2)			/* round up if necessary */
      r++;
    if (r > 63)				/* ensure in range 0..63 */
      r = 63;
    p++;			     	/* update pointer */

    g = *p >> 2;
    if ((*p&3) >= 2)
      g++;
    if (g > 63)
      g = 63;
    p++;

    b = *p >> 2;
    if ((*p&3) >= 2)
      b++;
    if (b > 63)
      b = 63;
    p++;

    _AX = 0x1010;			/* set up registers */
    _BX = count;
    _DH = r;
    _CH = g;
    _CL = b;
    geninterrupt(0x10);			/* call the video BIOS */
  }
}


/***
**** getBlockPalette
***/
void TGE_getBlockPalette(unsigned firstReg, unsigned numRegs, void far *data)
{
  register unsigned count;
  unsigned char r, g, b;
  unsigned char huge *p;

  p = (unsigned char*) data;		/* initialize pointer */

  for (count=firstReg; count<=numRegs; count++)	/* main loop */
  {
    _AX = 0x1015;			/* set up registers */
    _BX = count;
    geninterrupt(0x10);			/* call the video BIOS */
    r = _DH;
    g = _CH;
    b = _CL;

    *p = r << 2;			/* translate 6-bit to 8-bit */
    p++;				/* and store values */
    *p = g << 2;
    p++;
    *p = b << 2;
    p++;
  }
}


/***
**** clearGraphics
***/
void TGE_clearGraphics(unsigned colour)
{
  filledRect(0, 0, MAXX, MAXY, colour);
}


/***
**** ellipse
***/
void TGE_ellipse(int xc, int yc, int wide, int deep, unsigned colour)
{
  int x, y;
  long a, b;
  long Asquared, TwoAsquared;
  long Bsquared, TwoBsquared;
  long d, dx, dy;

  wide /= 2;
  deep /= 2;

  x = 0;				/* initialize variables */
  y = deep;
  a = wide;
  b = deep;
  Asquared = a * a;
  TwoAsquared = 2 * Asquared;
  Bsquared = b * b;
  TwoBsquared = 2 * Bsquared;

  d = Bsquared - Asquared*b + Asquared/4L;
  dx = 0;
  dy = TwoAsquared * b;

  while (dx < dy)
  {
    set4Pixels(x, y, xc, yc, colour);

    if (d > 0L)
    {
      y--;
      dy -= TwoAsquared;
      d -= dy;
    }

    x++;
    dx += TwoBsquared;
    d += Bsquared + dx;
  }

  d += (3L*(Asquared-Bsquared)/2L - (dx+dy)) / 2L;

  while (y >= 0)
  {
    set4Pixels(x, y, xc, yc, colour);

    if (d < 0L)
    {
      x++;
      dx += TwoBsquared;
      d += dx;
    }

    y--;
    dy -= TwoAsquared;
    d += Asquared - dy;
  }
}

void near set4Pixels(int x, int y, int xc, int yc, unsigned colour)
{
  register int xCoord, yCoord;

  xCoord = xc + x;
  yCoord = yc + y;
  if (clipPoint(xCoord, yCoord))
  {
    putPixel(xCoord, yCoord, colour);
  }

  xCoord = xc - x;
  if (clipPoint(xCoord, yCoord))
  {
    putPixel(xCoord, yCoord, colour);
  }

  xCoord = xc + x;
  yCoord = yc - y;
  if (clipPoint(xCoord, yCoord))
  {
    putPixel(xCoord, yCoord, colour);
  }

  xCoord = xc - x;
  if (clipPoint(xCoord, yCoord))
  {
    putPixel(xCoord, yCoord, colour);
  }
}


/***
**** filledEllipse
***/
void TGE_filledEllipse(int xc, int yc, int wide, int deep, unsigned colour)
{
  int x, y;
  long a, b;
  long Asquared, TwoAsquared;
  long Bsquared, TwoBsquared;
  long d, dx, dy;

  wide /= 2;
  deep /= 2;

  x = 0;				/* initialize variables */
  y = deep;
  a = wide;
  b = deep;
  Asquared = a * a;
  TwoAsquared = 2 * Asquared;
  Bsquared = b * b;
  TwoBsquared = 2 * Bsquared;

  d = Bsquared - Asquared*b + Asquared/4L;
  dx = 0;
  dy = TwoAsquared * b;

  while (dx < dy)
  {
    fill4Pixels(x, y, xc, yc, colour);

    if (d > 0L)
    {
      y--;
      dy -= TwoAsquared;
      d -= dy;
    }

    x++;
    dx += TwoBsquared;
    d += Bsquared + dx;
  }

  d += (3L*(Asquared-Bsquared)/2L - (dx+dy)) / 2L;

  while (y >= 0)
  {
    fill4Pixels(x, y, xc, yc, colour);

    if (d < 0L)
    {
      x++;
      dx += TwoBsquared;
      d += dx;
    }

    y--;
    dy -= TwoAsquared;
    d += Asquared - dy;
  }
}

void near fill4Pixels(int x, int y, int xc, int yc, unsigned colour)
{
  register int xCoord1, xCoord2, yCoord;
  int temp;

  xCoord1 = xc - x;
  xCoord2 = xc + x;

  if (xCoord1>VIEWPORTLRX || xCoord2<VIEWPORTULX)
    return;

  if (xCoord1 > xCoord2)		/* ensure xCoord1 <= xCoord2 */
  {
    temp = xCoord1;
    xCoord1 = xCoord2;
    xCoord2 = temp;
  }

  yCoord = yc - y;
  if (yCoord>=VIEWPORTULY && yCoord<=VIEWPORTLRY)
  {
    if (xCoord1 < VIEWPORTULX)		/* clip xCoord1 horizontally */
      xCoord1 = VIEWPORTULX;
    else if (xCoord1 > VIEWPORTLRX)
      xCoord1 = VIEWPORTLRX;
    if (xCoord2 < VIEWPORTULX)		/* clip xCoord2 horizontally */
      xCoord2 = VIEWPORTULX;
    else if (xCoord2 > VIEWPORTLRX)
      xCoord2 = VIEWPORTLRX;
    horizLine(yCoord, xCoord1, xCoord2, colour);
  }

  yCoord = yc + y;
  if (yCoord>=VIEWPORTULY && yCoord<=VIEWPORTLRY)
  {
    if (xCoord1 < VIEWPORTULX)		/* clip xCoord1 horizontally */
      xCoord1 = VIEWPORTULX;
    else if (xCoord1 > VIEWPORTLRX)
      xCoord1 = VIEWPORTLRX;
    if (xCoord2 < VIEWPORTULX)		/* clip xCoord2 horizontally */
      xCoord2 = VIEWPORTULX;
    else if (xCoord2 > VIEWPORTLRX)
      xCoord2 = VIEWPORTLRX;
    horizLine(yCoord, xCoord1, xCoord2, colour);
  }
}


/***
**** circle
***/
void TGE_circle(int x, int y, int radius, unsigned colour)
{
  ellipse(x, y, 2*radius, (2*radius*(4*YRATIO))/(3*XRATIO), colour);
}


/***
**** filledCircle
***/
void TGE_filledCircle(int x, int y, int radius, unsigned colour)
{
  filledEllipse(x, y, 2*radius, (2*radius*(4*YRATIO))/(3*XRATIO), colour);
}


/***
**** fillRegion
***/
static struct params {
  int x, y;
};
static unsigned regionColour, fillColour;
static int okToPush;
static struct params huge *stack;
static unsigned long stackSize, stackPtr;

void TGE_fillRegion(int seedX, int seedY, unsigned colour)
{
  struct params holder;
  register int x, v;
  int xl, xr, y;

  if (!clipPoint(seedX, seedY))			/* abort if seed point */
    return;					/* is outside viewport */

  regionColour = getPixel(seedX, seedY);	/* initialize variables */
  if (regionColour == colour)
    return;
  fillColour = colour;

  stackSize = farcoreleft();			/* grab a very large */
  (void *)stack = farmalloc(stackSize);		/*  stack buffer     */
  stackSize /= sizeof(struct params);
  stackPtr = 0;

  okToPush = 1;
  holder.x = seedX;			/* push initial parameters */
  holder.y = seedY;
  stack[stackPtr] = holder;
  stackPtr++;

  while (stackPtr && stackPtr<=stackSize)	/* main loop */
  {
    stackPtr--;
    holder = stack[stackPtr];

    xl = xr = holder.x;
    y = holder.y;

    x = xl;					/* scan left */
    if (x > VIEWPORTULX)
    {
      do
      {
        x--;
        v = getPixel(x, y);
      }
      while ((v==regionColour) && (v!=fillColour) && (x>=VIEWPORTULX));
      x++;
      xl = x;
    }

    x = xr;					/* scan right */
    if (x <= VIEWPORTLRX)
    {
      do
      {
        x++;
        v = getPixel(x, y);
      }
      while ((v==regionColour) && (v!=fillColour) && (x<=VIEWPORTLRX));
      x--;
      xr = x;
    }

    horizLine(y, xl, xr, fillColour);		/* fill the line */

    /* scan upwards */
    if (y > VIEWPORTULY)
    {
      for (x=xl; x<=xr; x++)
      {
        v = getPixel(x, y-1);
        if (v==regionColour && okToPush)
        {
          holder.x = x;
          holder.y = y-1;
          stack[stackPtr] = holder;
          stackPtr++;
          if (stackPtr != stackSize)
          {
            do
            {
              x++;
              v = getPixel(x, y-1);
            }
            while (v==regionColour && x<=xr);
            x--;
          }
          else
            okToPush = 0;
        }
      }
    }

    /* scan downwards */
    if (y < VIEWPORTLRY)
    {
      for (x=xl; x<=xr; x++)
      {
        v = getPixel(x, y+1);
        if (v==regionColour && okToPush)
        {
          holder.x = x;
          holder.y = y+1;
          stack[stackPtr] = holder;
          stackPtr++;
          if (stackPtr != stackSize)
          {
            do
            {
              x++;
              v = getPixel(x, y+1);
            }
            while (v==regionColour && x<=xr);
            x--;
          }
          else
            okToPush = 0;
        }
      }
    }
  };

  farfree((void far*)stack);			/* release stack memory */
}


/***
**** colourCloseTo
***/
static unsigned char palette[768];

unsigned TGE_colourCloseTo(unsigned char red, unsigned char green, unsigned
	char blue)
{
  register int count;
  int redDif, greenDif, blueDif, curDif, bestDif=1000, curNet, bestNet=1000;
  unsigned bestColour;

  getBlockPalette(0, 256, palette);		/* get the palette */

  for (count=255; count>=0; count--)
  {
    redDif = red - palette[3*count];	   /* calculate greatest difference */
    greenDif = green - palette[3*count+1];
    blueDif = blue - palette[3*count+2];
    curNet = redDif + greenDif + blueDif;
    if (redDif < 0)
      redDif = -redDif;
    if (greenDif < 0)
      greenDif = -greenDif;
    if (blueDif < 0)
      blueDif = -blueDif;
    curDif = max(max(redDif,greenDif), blueDif);

    if (!curDif)			/* if same colour, return */
    {
      bestColour = count;
      break;
    }
    if (curDif < bestDif)		/* if better colour, set bestColour */
    {
      if (curNet <= bestNet)
      {
        bestDif = curDif;
        bestColour = count;
      }
    }
  }

  return (bestColour);
}


/***
**** colourCloseToX
***/
unsigned TGE_colourCloseToX(unsigned char red, unsigned char green, unsigned
	char blue, unsigned colourExclude)
{
  register int count;
  int redDif, greenDif, blueDif, curDif, bestDif=1000, curNet, bestNet=1000;
  unsigned bestColour;

  getBlockPalette(0, 256, palette);		/* get the palette */

  for (count=255; count>=0; count--)
  {
    if (count != colourExclude)
    {
      redDif = red - palette[3*count];  /* calculate greatest difference */
      greenDif = green - palette[3*count+1];
      blueDif = blue - palette[3*count+2];
      curNet = redDif + greenDif + blueDif;
      if (redDif < 0)
        redDif = -redDif;
      if (greenDif < 0)
        greenDif = -greenDif;
      if (blueDif < 0)
        blueDif = -blueDif;
      curDif = max(max(redDif,greenDif), blueDif);

      if (!curDif)			/* if same colour, return */
      {
        bestColour = count;
        break;
      }
      if (curDif < bestDif)		/* if better colour, set bestColour */
      {
        if (curNet <= bestNet)
        {
          bestDif = curDif;
          bestColour = count;
        }
      }
    }
  }

  return (bestColour);
}


/*****
****** Viewport routines
*****/

void TGE_setViewport(int ulx, int uly, int lrx, int lry)
{
  disable();
  VIEWPORTULX = ulx;				/* set values */
  VIEWPORTULY = uly;
  VIEWPORTLRX = lrx;
  VIEWPORTLRY = lry;
  enable();
}


void TGE_getViewport(int far *ulx, int far *uly, int far *lrx, int far *lry)
{
  disable();
  *ulx = VIEWPORTULX;				/* get values */
  *uly = VIEWPORTULY;
  *lrx = VIEWPORTLRX;
  *lry = VIEWPORTLRY;
  enable();
}