/*****************************************************************************
*       The Graphics Engine version 1.20                                     *
*                                                                            *
*       The Graphics Engine code and documentation are Copyright (c) 1993    *
*	by Matthew Hildebrand.                                               *
*                                                                            *
*       Unauthorised usage or modification of any or all of The Graphics     *
*       Engine is strictly prohibited.                                       *
*****************************************************************************/

#ifndef TGEdotH
#define TGEdotH


/* The basic structure used to store function pointers and mode data */

struct GraphDrv {
  int far (*_initGraphics)(void);
  void far (*_deInitGraphics)(void);
  void huge (*_putImage)(int x, int y, void far *image);
  void huge (*_putImageInv)(int x, int y, void far *image);
  void huge (*_getImage)(int ulx, int uly, int lrx, int lry, void far *image);
  void far (*_putLine)(int lineNum, int xOff, int lineLen, void far *buf);
  void far (*_getLine)(int lineNum, int xOff, int lineLen, void far *buf);
  unsigned long far (*_imageSize)(int ulx, int uly, int lrx, int lry);
  void far (*_putPixel)(int x, int y, unsigned colour);
  unsigned far (*_getPixel)(int x, int y);
  void far (*_line)(int x1, int y1, int x2, int y2, unsigned colour);
  void far (*_horizLine)(int y, int x1, int x2, unsigned colour);
  void far (*_drawRect)(int ulx, int uly, int lrx, int lry, unsigned colour);
  void far (*_filledRect)(int ulx, int uly, int lrx, int lry, unsigned
  	colour);
  void far (*_setPaletteReg)(unsigned palReg, unsigned char red, unsigned
  	char green, unsigned char blue);
  void far (*_getPaletteReg)(unsigned palReg, unsigned char far *red,
  	unsigned char far *green, unsigned char far *blue);
  void far (*_setBlockPalette)(unsigned firstReg, unsigned numRegs, void
  	far *data);
  void far (*_getBlockPalette)(unsigned firstReg, unsigned numRegs, void
  	far *data);
  void far (*_clearGraphics)(unsigned colour);
  void far (*_ellipse)(int xc, int yc, int wide, int deep, unsigned
  	colour);
  void far (*_filledEllipse)(int xc, int yc, int wide, int deep, unsigned
  	colour);
  void far (*_circle)(int x, int y, int radius, unsigned colour);
  void far (*_filledCircle)(int x, int y, int radius, unsigned colour);
  void far (*_fillRegion)(int x, int y, unsigned colour);
  unsigned far (*_colourCloseTo)(unsigned char red, unsigned char green,
  	unsigned char blue);
  unsigned far (*_colourCloseToX)(unsigned char red, unsigned char green,
  	unsigned char blue, unsigned colourExclude);
  void far (*_slot1)(void);
  void far (*_slot2)(void);
  void far (*_slot3)(void);
  void far (*_slot4)(void);
  void far (*_slot5)(void);
  void far (*_slot6)(void);
  void far (*_slot7)(void);
  void far (*_slot8)(void);
  void far (*_slot9)(void);
  void far (*_slot10)(void);
  void far (*_slot11)(void);
  void far (*_slot12)(void);
  void far (*_slot13)(void);
  void far (*_slot14)(void);
  void far (*_slot15)(void);
  void far (*_slot16)(void);
  void far (*_slot17)(void);
  int maxx;				/* screen dimensions */
  int maxy;
  unsigned maxColour;			/* maximum colour number */
  int xRatio, yRatio;			/* screen aspect ratio */
  int viewportULX, viewportULY;		/* upper-left viewport coords */
  int viewportLRX, viewportLRY;		/* lower-right viewport coords */
};


/* Note that this C/C++ bit might not work properly with your compiler.  If
   you are using C++ and your linker reports that these functions are not
   found, try explicitly forcing the extern "C" syntax by removing these
   preprocessor directives. */

#ifdef __cplusplus
extern "C" {
#endif

/* load/unload routines */
int loadGraphDriver(char *filename);	/* use extern "C" syntax if C++ */
void unloadGraphDriver(void);

/* viewport routines */
void TGE_setViewport(int ulx, int uly, int lrx, int lry);
void TGE_getViewport(int far *ulx, int far *uly, int far *lrx, int far *lry);

/* clipping routines */
int TGE_clipLine(int *x1, int *y1, int *x2, int *y2);
int TGE_clipFilledRect(int *ulx, int *uly, int *lrx, int *lry);

#ifdef __cplusplus
};
#endif


/* Declare the structure used by the system so that programs can access it. */

extern struct GraphDrv far *TGEsys;


/* The function names of TGE's functions are defined here.  Since they
   are really just macros pretending to be functions, you can easily
   change their names here if you don't like the default ones. */

#define	initGraphics		(TGEsys->_initGraphics)
#define	deInitGraphics		(TGEsys->_deInitGraphics)
#define	putImage		(TGEsys->_putImage)
#define putImageInv		(TGEsys->_putImageInv)
#define	getImage		(TGEsys->_getImage)
#define putLine			(TGEsys->_putLine)
#define getLine			(TGEsys->_getLine)
#define	imageSize		(TGEsys->_imageSize)
#define	putPixel		(TGEsys->_putPixel)
#define	getPixel		(TGEsys->_getPixel)
#define	line			(TGEsys->_line)
#define horizLine		(TGEsys->_horizLine)
#define	drawRect		(TGEsys->_drawRect)
#define	filledRect		(TGEsys->_filledRect)
#define	setPaletteReg		(TGEsys->_setPaletteReg)
#define	getPaletteReg		(TGEsys->_getPaletteReg)
#define	setBlockPalette		(TGEsys->_setBlockPalette)
#define	getBlockPalette		(TGEsys->_getBlockPalette)
#define	clearGraphics		(TGEsys->_clearGraphics)
#define ellipse			(TGEsys->_ellipse)
#define filledEllipse		(TGEsys->_filledEllipse)
#define circle			(TGEsys->_circle)
#define filledCircle		(TGEsys->_filledCircle)
#define fillRegion		(TGEsys->_fillRegion)
#define colourCloseTo		(TGEsys->_colourCloseTo)
#define colourCloseToX		(TGEsys->_colourCloseToX)

#define setViewport		TGE_setViewport		/* viewports */
#define getViewport		TGE_getViewport

#define clipLine		TGE_clipLine		/* clipping */
#define clipFilledRect		TGE_clipFilledRect
#define clipPoint(x,y)		TGE_clipPoint(x,y)
#define pointOnScreen(x,y)	TGE_pointOnScreen(x,y)


/* These macros provide easy access to important mode information in the
   TGE system structure. */

#define MAXX			(TGEsys->maxx)
#define MAXY			(TGEsys->maxy)
#define MAXCOLOUR		(TGEsys->maxColour)
#define XRATIO			(TGEsys->xRatio)
#define YRATIO			(TGEsys->yRatio)
#define VIEWPORTULX		(TGEsys->viewportULX)
#define VIEWPORTULY		(TGEsys->viewportULY)
#define VIEWPORTLRX		(TGEsys->viewportLRX)
#define VIEWPORTLRY		(TGEsys->viewportLRY)


/* These macros are standard with Borland C/C++ compilers; this bit defines
   them if they aren't already defined. */

#ifndef MK_FP
#define MK_FP(seg,ofs)((void _seg *)(seg)+(void near *)(ofs))
#endif
#ifndef FP_SEG
#define FP_SEG(fp)((unsigned)(void _seg *)(void far *)(fp))
#endif
#ifndef FP_OFF
#define FP_OFF(fp)((unsigned)(fp))
#endif


/* These macros are the return codes from loadGraphDriver() */

#define TGE_SUCCESS	 0		/* driver loaded successfully */
#define TGE_OPEN_ERR	-1		/* error opening given file */
#define TGE_FORMAT_ERR	-2		/* file is not a TGE driver */
#define TGE_ALLOC_ERR	-3	    	/* error allocating memory */
#define TGE_FILE_ERR	-4		/* file I/O error */


/******
******* Point clipping routines
******/

/* Return true if a point is on-screen, or 0 if it isn't */
#define TGE_pointOnScreen(x,y) \
  ((x>=0) && (x<=MAXX) && (y>=0) && (y<=MAXY))

/* Return true if a point is within the current viewport, or 0 if it isn't */
#define TGE_clipPoint(x,y) \
  ((x>=VIEWPORTULX) && (x<=VIEWPORTLRX) && \
  (y>=VIEWPORTULY) && (y<=VIEWPORTLRY))

#endif